/* xenv - expand environment variables in input files  -*- c -*-
   Copyright (C) 2021 Sergey Poznyakoff

   Xenv is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3 of the License, or (at your
   option) any later version.

   Xenv is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with xenv. If not, see <http://www.gnu.org/licenses/>. */
%{
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <sysexits.h>
#include <assert.h>

char const *progname;       /* Program name */
char *filename;             /* Name of the input file. */
size_t filename_size;       /* Size of memory allocated for filename. */
unsigned lineno;            /* Line number in the input file. */
unsigned start_line = 1;    /* Line where the last state change occurred. */
int undef_error_option;     /* Treat undefined variables as error. */
int synclines_option;       /* Generate `#line NUM "FILE"' lines. */
int retain_unexpanded_option;/* Retain unexpanded constructs in the output. */
int status = 0;             /* Exit status */
int sigil = '$';            /* Current sigil. */ 
char const sigil_chars[] = "$@%&#";
                            /* Characters that are allowed for use as a sigil.
			       Note, that these are duplicated in the SPECIAL
			       character class definition below. */

static int bracecount = 0;  /* Curly brace nesting level. */
static int emit_syncline;

static void push_state(int newstate);
static void pop_state(void);

static char *findenv(char const *ident, int len);
void expandenv(char const *ident, int len);

enum {  /* Variable expansion modifiers. */
	EXP_DEFVAL = '-',   /* Use default value. */
	EXP_ASSIGN = '=',   /* Assign default value. */
	EXP_ERROR  = '?',   /* Display error if null or unset. */
	EXP_ALTER  = '+',   /* Use alternate value. */
	EXP_TERNARY = '|'   /* Ternary operator. */
};

static void expand_inline_push(int type, char *varname, int varlen, int suppress);
static int expand_inline_pop(void);
static void expand_inline_flop(void);

static int cond_peek(unsigned *);
static void cond_push(int result);
static int cond_pop(void);

static int parse_line_directive(char *text);

static void echo(char const*, size_t);
#define ECHO echo(yytext, yyleng)


/*
 * TEXT is a portion of preprocessor statement after double sigil marker.
 * Find the variable identifier part.  Return a pointer to the identifier
 * and store its length in LEN.
 * If KW is supplied, initialize it with the pointer to the preprocessor
 * keyword (after $$ and optional whitespace).
 */
static char *
find_ident(char *text, int *len, char **kw)
{
	if (text[0] == ' ' || text[0] == '\t')
		text += strspn(text, " \t");
	if (kw)
		*kw = text;
	text += strcspn(text, " \t");
	text += strspn(text, " \t");
	*len = strcspn(text, " \t\n");
	return text;
}

/*
 * Memory management with error handling.
 */
static void
enomem(void)
{
	fprintf(stderr, "%s: not enough memory\n", progname);
	abort();
}

char *
xmalloc(size_t size)
{
	char *p = malloc(size);
	if (!p)
		enomem();
	return p;
}

char *
xstrdup(char *str)
{
	char *p = strdup(str);
	if (!p)
		enomem();
	return p;
}

char *
xrealloc(char *ptr, size_t size)
{
	char *p = realloc(ptr, size);
	if (!p)
		enomem();
	return p;
}
	
static void
set_filename(char const *name, size_t len)
{
	if (len + 1 > filename_size) {
		filename_size = len + 1;
		filename = xrealloc(filename, filename_size);
	}
	memcpy(filename, name, len);
	filename[len] = 0;
}

/*
 * General-purpose stack.
 */
typedef struct xenv_stack {
	size_t entry_size;
	char *stack;
	size_t max_stack;
	size_t tos;
} XENV_STACK;

#define XENV_STACK_INITIALIZER(type)  { sizeof(type), NULL, 0, 0 }
#define XENV_INIT_SIZE 64

static inline int
xenv_stack_size(XENV_STACK *stk)
{
	return stk->tos;
}

static void *
xenv_stack_peek(XENV_STACK *stk, int dec)
{
	if (-dec > stk->tos)
		return NULL;
	return stk->stack + stk->entry_size * (stk->tos + dec);
}

static void *
xenv_stack_push(XENV_STACK *stk)
{
	int n = stk->max_stack;
	if (stk->tos == stk->max_stack) {
		if (!stk->stack)
			n = XENV_INIT_SIZE;
		else if (((size_t) -1) / 3 * 2 / stk->entry_size <= n)
			enomem();
		else
			n += (n + 1) / 2;
		stk->stack = xrealloc(stk->stack, n * stk->entry_size);
		stk->max_stack = n;
	}
	++stk->tos;
	return xenv_stack_peek(stk, -1);
}

static void *
xenv_stack_pop(XENV_STACK *stk)
{
	if (stk->tos == 0) {
		fprintf(stderr, "%s:%u: out of popup space\n",
			filename, lineno);
		exit(EX_SOFTWARE);
	}
	--stk->tos;
	return xenv_stack_peek(stk, 0);
}

%}
  /* Identifier */
IDENT [a-zA-Z_][a-zA-Z_0-9]*
  /* Whitespace */
WS [ \t][ \t]*
  /* Optional whitespace */
OWS [ \t]*
  /* Special character: can be a sigil */
SPECIAL [$@%&#]

  /*
   * States
   */
%x COMMENT FALSE SIGIL DSIGIL VERBATIM
  /*
   * Possible state transitions are:
   *
   *  State     Input         New state          Comment
   *  --------  ------------  ----------------   -------
   *   INITIAL  <sigil>       SIGIL
   *   SIGIL    {*            COMMENT
   *   SIGIL    <sigil>       DSIGIL
   *   COMMENT  *}            INITIAL
   *   DSIGIL   "verbatim"    VERBATIM           Unless prior state was FALSE.
   *   DSIGIL   "verbatim"    FALSE              If prior state was FALSE.
   *   DSIGIL   "ifset"       INITIAL or FALSE   Depending on condition.
   *            "ifnset"                         --"--
   *            "ifdef"                          --"--
   *            "ifndef"                         --"--
   *            "else"                           --"--
   *   VERBATIM "end"         INITIAL
   *   FALSE    "else"        INITIAL 
   *            "endif"
   */
%%
<INITIAL>{
\n+  {
    ECHO;
    lineno += yyleng;
}
\\. {
	if (yytext[1] == '\\' || yytext[1] == sigil)
		fputc(yytext[1], yyout);
	else
		ECHO;
 }
\{  { bracecount++; ECHO; }
\}  { if (expand_inline_pop())
              ECHO;
      bracecount--;
 }
\|  { expand_inline_flop(); }

^"#line ".*\n {
	if (parse_line_directive(yytext)) {
		ECHO;
		lineno++;
	}
 }

{SPECIAL} {
	   if (yytext[0] == sigil)		
  	        push_state(SIGIL);
	   else
  	        ECHO; 	
         }

^{OWS}{SPECIAL}{2} {
	if (yytext[yyleng-2] == sigil) {
		if (yytext[yyleng-1] == sigil) {
			push_state(DSIGIL);
		} else {
			yyless(yyleng-1);
			echo(yytext, yyleng-2);
			push_state(SIGIL);
		}
	} else {
		yyless(yyleng-1);
		echo(yytext, yyleng-1);
	}
 }	
}

<COMMENT>{
[^*\n]*        /* eat anything that's not a '*' */
"*"+[^*}\n]*   /* eat up '*'s not followed by '}'s */
\n             lineno++; /* Keep track of line numbers. */
"*"+"}"        { pop_state(); emit_syncline = 1; }
}

<SIGIL>{
"{*" { pop_state(); push_state(COMMENT); }
"(" {
	int nesting = 1;
	unsigned line = start_line; /* Save start_line for eventual use in
				       the diagnostic message below */
	int c;
	while ((c = input()) != EOF) {
                if (c == '(')
		        nesting++;
		else if (c == ')') {
	                if (--nesting == 0)
				break;
                } else if (c == '\n')
			lineno++;
	        fputc(c, yyout);   
        }
	pop_state();
	if (c == EOF) {
		fprintf(stderr, "%s:%u: end of file in inline verbatim text\n",
			filename, line);
		return EOF;
	}
 }
{IDENT} { expandenv(yytext, yyleng);
	  pop_state(); }
\{{IDENT}\} { expandenv(yytext + 1, yyleng - 2);
	      pop_state(); }
\{{IDENT}:?[-=?+|] {
	int len;
	char *val;
	int test_null = 0;
	int type = yytext[yyleng-1];
	int suppress;
	
	pop_state();
 	if (yytext[yyleng-2] == ':') {
		test_null = 1;
		len = yyleng - 3;
	} else
		len = yyleng - 2;
	val = findenv(yytext + 1, len);
	if (val && !(test_null && *val == 0)) {
	        if (type == EXP_TERNARY || type == EXP_ALTER) {
	                suppress = 0;
                } else {
		        fwrite(val, strlen(val), 1, yyout);
		        suppress = 1;
                }
        } else if (type == EXP_TERNARY || type == EXP_ALTER) {
	        suppress = 1;
        } else {
	        suppress = 0;
		if (type == EXP_ERROR ) {
	                int c = input();
			if (c == '}') {
	                        /* Provide default error message */
	                        if (test_null)
				        fprintf(stderr, "%s:%u: variable %.*s null or not set\n",
		                                filename, lineno, len, yytext + 1);
			        else
				        fprintf(stderr, "%s:%u: variable %.*s not set\n",
		                                filename, lineno, len, yytext + 1);
			        YY_BREAK;
                         }
	                 unput(c);
                }
        }
	bracecount++;
	expand_inline_push(type, yytext + 1, len, suppress);
}

.   {   
	char s[2];
	s[0] = sigil;
	s[1] = yytext[0];
	echo(s, 2);
	pop_state();
 }
}

<DSIGIL>{
{OWS}"verbatim"{OWS}"\n" {
	pop_state();
	if (YYSTATE != FALSE)
		push_state(VERBATIM);
	lineno++;
 }

{OWS}"sigil"{WS}{SPECIAL}{OWS}"\n" {
	pop_state();
	if (YYSTATE != FALSE) {
		int len;
		char *ident = find_ident(yytext, &len, NULL);
		sigil = *ident;
	}
	lineno++;
 }

{OWS}"unset"{WS}{IDENT}{OWS}\n {
	int len;
	char *ident = find_ident(yytext, &len, NULL);

	ident[len] = 0;
	unsetenv(ident);

	lineno++;
	emit_syncline = 1;
	pop_state();
 }

{OWS}"set"{WS}{IDENT}{OWS}\n {
	int len;
	char *ident = find_ident(yytext, &len, NULL);

	ident[len] = 0;
	setenv(ident, "", 1);

	lineno++;
	emit_syncline = 1;
	pop_state();
 }

{OWS}"set"{WS}{IDENT}{WS}\".*\"{OWS}\n {
	int len;
	char *ident = find_ident(yytext, &len, NULL);
	char *val, *p, *q;

	ident[len] = 0;
	val = strchr(ident + len + 1, '"') + 1;
	p = q = val;
	while (*q != '"') {
		if (*q == '\\') {
			if (!*++q)
				break;
		}
		*p++ = *q++;
	}
	*p = 0;
	setenv(ident, val, 1);

	lineno++;
	emit_syncline = 1;
	pop_state();
 }

{OWS}"if"n?(def|set){WS}{IDENT}{OWS}\n {
	int result;
	
	pop_state();
	result = cond_peek(NULL);
	if (result) {
		int len;
		char *ident;
		char *val;
		char *kw;
		int neg;
		
		ident = find_ident(yytext, &len, &kw);
		neg = kw[2] == 'n';
		val = findenv(ident, len);
		result = val != NULL && (kw[2+neg] == 'd' || *val != 0);
		if (neg)
			result = !result;
        }
	cond_push(result);
	lineno++;
 }

{OWS}"else"{OWS}\n {
	pop_state();
	if (cond_peek(NULL) == -1) {
		fprintf(stderr, "%s:%u: $$else without matching $$if\n",
			filename, lineno);
		ECHO;
	} else {	
		int result = cond_pop();
		if (cond_peek(NULL))
			result = !result;
		cond_push(result);
	}
	lineno++;
}

{OWS}"endif"{OWS}\n {
	pop_state();
	if (cond_peek(NULL) == -1) {
		fprintf(stderr, "%s:%u: $$endif without matching $$if\n",
			filename, lineno);
		ECHO;
	} else {
		cond_pop();
	}
	lineno++;
	emit_syncline = 1;
 }

. {
	pop_state();
	if (YYSTATE != FALSE) {
		char s[3];
		s[0] = s[1] = sigil;
		s[2] = yytext[0];
		fprintf(stderr, "%s:%u: unrecognized double-sigil statement (%.3s)\n",
			filename, lineno, s);
		echo(s, 3);
	}
 }
}

<FALSE>{
\n+     lineno += yyleng;
^{OWS}{SPECIAL}{2} {
	if (yytext[yyleng-2] == sigil && yytext[yyleng-1] == sigil) {
		push_state(DSIGIL);
	}
 }	
. ;
}

<VERBATIM>{
\n+    { ECHO; lineno += yyleng; }
^{OWS}{SPECIAL}{2}{OWS}"end"{OWS}"\n" {
	if (yytext[0] == sigil && yytext[1] == sigil) {
		pop_state();
		lineno++;
	} else {
		ECHO;
	}
 }
}
%%
static void
echo(char const *text, size_t len)
{
	if (synclines_option && emit_syncline && text[0] == '\n'
	    && YYSTATE == INITIAL) {
		fputc(text[0], yyout);
		fprintf(yyout, "#line %u \"%s\"\n", lineno, filename);
		text++;
		len--;
		emit_syncline = 0;
	}
	if (len)
		fwrite(text, len, 1, yyout);
}


extern char **environ;

/*
 * Same as getenv(3), but uses first LEN bytes of IDENT as the name of
 * a variable.
 */
static char *
findenv(char const *ident, int len)
{
	size_t i;
	for (i = 0; environ[i]; i++) {
		size_t j;
		for (j = 0; j < len; j++)
			if (environ[i][j] != ident[j])
				break;
		if (j == len && environ[i][j] == '=')
			return environ[i] + j + 1;
	}
	return NULL;
}

/*
 * Expand environment variable whose name is given by first LEN bytes of
 * IDENT.
 */
void
expandenv(char const *ident, int len)
{
	char *val = findenv(ident, len);
	
	if (val) {
		fwrite(val, 1, strlen(val), yyout);
		if (strchr(val, '\n'))
			emit_syncline = 1;
	} else {
		if (undef_error_option) {
			fprintf(stderr, "%s:%u: variable %.*s not defined\n",
				filename, lineno, len, ident);
			status = EX_DATAERR;
		}
		if (retain_unexpanded_option) {
			char c = sigil;
			echo(&c, 1);
			ECHO;
		}
	}
}	

struct saved_state {
	int state;
	unsigned line;
};

static XENV_STACK state_stack = XENV_STACK_INITIALIZER(struct saved_state);

static void
push_state(int newstate)
{
	struct saved_state *st = xenv_stack_push(&state_stack);
	st->state = YYSTATE;
	st->line = start_line;
	start_line = lineno;
	BEGIN(newstate);
}

static void
pop_state(void)
{
	struct saved_state *st = xenv_stack_pop(&state_stack);
	BEGIN(st->state);
	start_line = st->line;
}

struct expand_inline {
	int type;      /* Expansion type (one of the EXP_ constants above). */
	int suppress;  /* Suppress output. */
	int bracecount;/* Brace nesting level this expansion was created at. */
	FILE *fp;      /* Previous value of yyout */
	union {
		char *varname; /* type == EXP_ASSIGN - assign the expansion
				  to that variable */
		struct {       /* type == EXP_ERROR - location of the error */
			char *filename;
			unsigned lineno;
		} loc;
	};
};

static XENV_STACK expand_inline_stack = XENV_STACK_INITIALIZER(struct expand_inline);

/*
 * Push on stack the inline expansion entry of given TYPE.
 * VARNAME and VARLEN identify the variable if type is EXP_ASSIGN.
 * SUPPRESS is 1 if the output is to be suppressed and 0 otherwise.
 */
static void
expand_inline_push(int type, char *varname, int varlen, int suppress)
{
	struct expand_inline *ep;
	
	ep = xenv_stack_push(&expand_inline_stack);
	ep->type = type;
	if (type == EXP_ASSIGN) {
		assert(varname && varlen);
		ep->varname = xmalloc(varlen + 1);
		memcpy(ep->varname, varname, varlen);
		ep->varname[varlen] = 0;
        } else if (type == EXP_ERROR) {
  	        ep->loc.filename = xstrdup(filename);
		ep->loc.lineno = lineno;
        }
	ep->fp = yyout;
	ep->suppress = suppress;
	ep->bracecount = bracecount;
	if (suppress) {
		yyout = fopen("/dev/null", "w+b");
		if (!yyout) {
			fprintf(stderr, "%s: can't open /dev/null: %s\n",
				progname, strerror(errno));
			exit(EX_OSERR);
		}
	} else if ((yyout = tmpfile()) == NULL) {
		fprintf(stderr, "%s: can't create temporary file: %s\n",
			progname, strerror(errno));
		exit(EX_OSERR);
	}
}

/*
 * Flop the suppression state of the ternary operator on the top of stack.
 * If there's no ternary on tos or its brace count doesn't match the current
 * one, echo yytext.
 */
static void
expand_inline_flop(void)
{
	struct expand_inline *ep = xenv_stack_peek(&expand_inline_stack, -1);
	if (ep &&
	    ep->bracecount == bracecount &&
	    ep->type == EXP_TERNARY) {
		int suppress = ep->suppress;
		expand_inline_pop();
		expand_inline_push(EXP_TERNARY, NULL, 0, !suppress);
	} else
		ECHO;  
}

/*
 * Pop the topmost expansion entry off the stack.
 * Return 1 if the operation is not applicable (wrong state or brace level
 * mismatch).
 * Return 0 on success.
 */
static int
expand_inline_pop(void)
{
	struct expand_inline *ep;

	ep = xenv_stack_peek(&expand_inline_stack, -1);
	if (ep == NULL || ep->bracecount != bracecount) {
		return 1;
	}
	xenv_stack_pop(&expand_inline_stack);
	if (!ep->suppress) {
		if (ep->type == EXP_ASSIGN) {
			long size;
			char *value;
			size_t n;
			
			fseek(yyout, 0, SEEK_END);
			size = ftell(yyout);
			value = xmalloc(size + 1);
			rewind(yyout);
			n = fread(value, 1, size, yyout);
			if (n == -1) {
				fprintf(stderr, "%s: read error: %s\n",
					progname, strerror(errno));
				exit(EX_SOFTWARE);
			} else if (n > 0 && n < size) {
				fprintf(stderr, "%s: short read\n", progname);
				exit(EX_SOFTWARE);
			}				
			value[n] = 0;
			setenv(ep->varname, value, 1);
			fwrite(value, size, 1, ep->fp);
			if (strchr(value, '\n'))
				emit_syncline = 1;
                } else if (ep->type == EXP_ERROR) {
	                int c;

			fprintf(stderr, "%s:%u: ", ep->loc.filename,
				ep->loc.lineno);
			free(ep->loc.filename);
			rewind(yyout);
			while ((c = fgetc(yyout)) != EOF)
				fputc(c, stderr);
			fputc('\n', stderr);
		} else {
			int c;
			
			rewind(yyout);
			while ((c = fgetc(yyout)) != EOF) {
				fputc(c, ep->fp);
				if (c == '\n')
					emit_syncline = 1;
			}
		}
	} else
		/* FIXME: Emit syncline only if there were newlines in the
		   suppressed text. */
		emit_syncline = 1;
	
	if (ep->type == EXP_ASSIGN)
		free(ep->varname);
	fclose(yyout);
	yyout = ep->fp;
	return 0;
}

/*
 * Condition stack.  Implements conditional preprocessor blocks like:
 *
 *   $$ifset
 *    TEXT1
 *   $$else
 *    TEXT2
 *   $$endif
 */

struct cond_state {
	int result;
	unsigned line;
};

static XENV_STACK cond_stack = XENV_STACK_INITIALIZER(struct cond_state);

/*
 * Return the evaluation result from the top of stack.
 */
static int
cond_peek(unsigned *line)
{
	struct cond_state *st = xenv_stack_peek(&cond_stack, -1);
	if (st == NULL)
		return -1;
	if (line)
		*line = st->line;
	return st->result;
}

/*
 * Push RESULT on stack. Switch to state FALSE if result is false. Otherwise,
 * request a syncline on output.
 */
static void
cond_push(int result)
{
	struct cond_state *st = xenv_stack_push(&cond_stack);
	st->result = result;
	st->line = lineno;
	if (!result)
		push_state(FALSE);
	else 
		emit_syncline = 1;
}

/*
 * Pop the entry off the stack.  If it is false, restore previous scanner
 * state.  Return the evaluation result.
 */
static int
cond_pop(void)
{
	struct cond_state *st = xenv_stack_pop(&cond_stack);
	if (!st->result)
		pop_state();
	return st->result;
}

static int
parse_line_directive(char *text)
{
	unsigned long n;
	char *fname, *p;
	
	text += sizeof("#line ")-1;

	n = strtoul(text, &p, 10);
	if ((n == 0 && errno == ERANGE) || n > INT_MAX ||
	    !(*p == 0 || *p == ' ' || *p == '\t' || *p == '\n'))
		return 1;
	while (*p == ' ' || *p == '\t')
		p++;
	if (*p == '\n' || *p == 0) {
		lineno = (int) n;
		return 0;
	}
	if (*p != '"')
		return 1;
	p++;
	fname = p;
	p = strchr(fname, '"');
	if (!p)
		return 1;
	set_filename(fname, p - fname);
	lineno = n;
	return 0;
}

/*
 * Input file handling.
 */
static char **input_files; /* Array of input files. */
static int input_index;    /* Index of the next file in input_files. */

static void
open_input(char const *name)
{
	FILE *fp;

	if (name[0] == '-' && name[1] == 0) {
		name = "<stdin>";
		fp = stdin;
	} else if ((fp = fopen(name, "r")) == NULL) {
		fprintf(stderr, "%s: can't open %s\n", progname, name);
		exit(errno == ENOENT ? EX_NOINPUT : EX_OSERR);
	}
	yyin = fp;
	set_filename(name, strlen(name));
	lineno = 1;
	if (synclines_option)
		fprintf(yyout, "#line %u \"%s\"\n", lineno, filename);
}

int
yywrap(void)
{
	if (yyin != stdin)
		fclose(yyin);
	if (input_files[input_index]) {
		open_input(input_files[input_index++]);
		return 0;
	}
	return 1;
}

void
usage(FILE *fp)
{
	fprintf(fp, "usage: %s [-hnsu] [-D NAME[=VALUE]] [-U NAME] FILE [FILE...]\n", progname);
	fprintf(fp, "expands environment variables in input files\n");
	fprintf(fp, "\nOPTIONS are:\n\n");
	fprintf(fp, "  -D NAME[=VALUE]  set environment variable NAME\n");
	fprintf(fp, "  -U NAME          unset environment variable NAME\n");
	fprintf(fp, "  -n               don't produce output, only report errors\n");
	fprintf(fp, "  -r               Retain unexpanded constructs in the output\n");
	fprintf(fp, "  -S CHAR          change sigil character (allowed chars: %s)\n", sigil_chars);
	fprintf(fp, "  -s               generate `#line NUM \"FILE\"' lines\n");
	fprintf(fp, "  -u               treat unset variables as errors\n");
	fprintf(fp, "  -v               print the version number and exit\n");
	fprintf(fp, "  -h               print this help text\n");
	fprintf(fp, "\n");
	fprintf(fp, "Report bugs and suggestions to <gray@gnu.org>\n");
}

static int copyright_year = 2021;

void
version(void)
{
	printf("%s (%s) %s\n", progname, PACKAGE_NAME, PACKAGE_VERSION);
	printf("Copyright (C) %d Sergey Poznyakoff\n", copyright_year);
	printf("\
License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>\nThis is free software: you are free to change and redistribute it.\n\
There is NO WARRANTY, to the extent permitted by law.\n\
");
}

int
main(int argc, char **argv)
{
	int c;
	int dry_run = 0;
	
	progname = argv[0];
	while ((c = getopt(argc, argv, "D:hnrS:sU:uv")) != EOF) {
		switch (c) {
		case 'D':{
			char *p = strchr(optarg, '=');
			if (p)
				*p++ = 0;
			else
				p = "";
			setenv(optarg, p, 1);
			break;
		}

		case 'U':
			unsetenv(optarg);
			break;
			
		case 'h':
			usage(stdout);
			return 0;

		case 'n':
			dry_run = 1;
			break;

		case 'r':
			retain_unexpanded_option = 1;
			break;

		case 'S':
			if (optarg[1]) {
				fprintf(stderr, "%s: sigil too long\n",
					progname);
				return EX_USAGE;
			}
			if (!strchr(sigil_chars, optarg[0])) {
				fprintf(stderr, "%s: %s is not allowed as a sigil; allowed characters are: %s\n",
					progname, optarg, sigil_chars);
				return EX_USAGE;
			}
			sigil = optarg[0];
			break;
			
		case 's':
			synclines_option = 1;
			break;
			
		case 'u':
			undef_error_option = 1;
			break;

		case 'v':
			version();
			return 0;
			
		default:
			usage(stderr);
			return EX_USAGE;
		}
	}

	if (dry_run) {
		int fd;
		
		if ((fd = open("/dev/null", O_WRONLY)) == -1 ||
		    dup2(fd, 1) == -1) {
			fprintf(stderr, "%s: can't open /dev/null for writing: %s\n",
				progname, strerror(errno));
			return EX_OSERR;
		}
		close(fd);
	}

	input_files = argv + optind;
	input_index = 0;
	if (input_files[0] == NULL) {
		static char *stdin_file[] = { "-", NULL };
		input_files = stdin_file;
	}
	yyout = stdout;
	open_input(input_files[input_index++]);
	yylex();

	switch (YYSTATE) {
	case INITIAL: {
		unsigned line;
		if (cond_peek(&line) != -1)
			fprintf(stderr, "%s:%u: end of file in conditional\n",
				filename, line);
		break;
	}
		
	case DSIGIL:
		fputc(sigil, yyout);
	case SIGIL:
		fputc(sigil, yyout);
		break;
		
	case COMMENT:
		fprintf(stderr, "%s:%u: end of file in comment\n",
			filename, start_line);
		break;

	case FALSE:
		fprintf(stderr, "%s:%u: end of file in conditional\n",
			filename, start_line);
		break;
		
	case VERBATIM:
		fprintf(stderr, "%s:%u: end of file in verbatim block\n",
			filename, start_line);
		break;
	}
	return status;
}

